/******************************************************************************
* LockState v0.1a
* Copyright © SFR 2017
* GNU GPL v3 applies (http://www.gnu.org/licenses/gpl.html)
* 
* GNU GPL v2 applies (https://www.gnu.org/licenses/gpl-2.0.html)
* ' a few mods by vovchik for inline svg icons
* Disclaimer: my 2nd C/GTK+ app, slightly better than previous,
* but still quite messy and I'm sure could've be done better...
******************************************************************************/

/* Build with:
     gcc lockstate-mod.c -o lockstate-mod `pkg-config --cflags --libs gtk+-2.0` -lX11 -O3 -s
	
*/

/* Core */
#include <X11/Xlib.h>
#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
/* gettext */
#include <libintl.h>
#include <locale.h>
#define _(STRING) gettext(STRING)

/* ========================================================================= */	

FILE *fp;
gchar filename[100];
const gchar appname[]		= "LockState v0.1";
const gchar settings[]		= "lockstate.conf";
char* svg = "<svg height='32' width='32' viewBox='0 0 16 16'> \
 <radialGradient id='rg' \
  fx='50%' fy='50%' cx='50%' cy='50%' r='50%'> \
   <stop offset='0%'   stop-color='white' stop-opacity='1'/> \
   <stop offset='100%' stop-color='#bbbbbb' stop-opacity='1' /> \
 </radialGradient> \
<rect width='16' height='16' x='0' y='0' rx='3' ry='4' \
fill='none' stroke-opacity='0.7' stroke-width='2'/> \
<rect width='14' height='14' x='1' y='1' rx='2' ry='3' \
fill='#999999'/> \
<rect width='14' height='13' x='1' y='1' rx='2' ry='3' \
fill='url(#rg)'/>  \
<g font-size='11' font-style='normal' font-variant='normal' \
font-weight='normal' font-stretch='normal' text-align='center' \
line-height='125%' letter-spacing='0' word-spacing='0' \
text-anchor='middle' fill='red' fill-opacity='1' \
stroke='none' font-family='sans' dominant-baseline='central'> \
<text x='50%' dy='11'>a</text> \
</g> \
</svg>";

/* Defaults */
gboolean state_caps		= TRUE;
gboolean state_num		= TRUE;
gboolean state_scroll	= FALSE;

/* Status icons */
gchar caps_on[]		= "/tmp/caps-on.svg";
gchar caps_off[]	= "/tmp/caps-off.svg";
gchar num_on[]		= "/tmp/num-on.svg";
gchar num_off[]		= "/tmp/num-off.svg";
gchar scroll_on[]	= "/tmp/scr-on.svg";
gchar scroll_off[]	= "/tmp/scr-off.svg";

GtkStatusIcon	*icon_caps;
GtkStatusIcon	*icon_num;
GtkStatusIcon	*icon_scroll;

/* ------------------------------------------------------------------------- */


/*Function to replace a string with another string*/

char *rep_str(const char *s, const char *old, const char *new1)
{
    char *ret;
    int i, count = 0;
    int newlen = strlen(new1);
    int oldlen = strlen(old);
    for (i = 0; s[i] != '\0'; i++)    
    {
        if (strstr(&s[i], old) == &s[i]) 
        {
            count++;
            i += oldlen - 1;
        }
    }
    ret = (char *)malloc(i + count * (newlen - oldlen));
    if (ret == NULL)
        exit(EXIT_FAILURE);
    i = 0;
    while (*s)
    {
        if (strstr(s, old) == s) //compare the substring with the newstring
        {
            strcpy(&ret[i], new1);
            i += newlen; //adding newlength to the new string
            s += oldlen;//adding the same old length the old string
        }
        else
        ret[i++] = *s++;
    }
    ret[i] = '\0';
    return ret;
}

void save_config (void) {

	gchar		*filename;
	GKeyFile	*config;
	GError		*error = NULL;
	gsize		length;
	gchar		*contents;
	
	filename = g_strconcat (g_get_user_config_dir(), "/", settings, NULL);
	
	config = g_key_file_new();

	g_key_file_set_boolean (config, "Settings", "CapsLock",		state_caps);
	g_key_file_set_boolean (config, "Settings", "NumLock",		state_num);
	g_key_file_set_boolean (config, "Settings", "ScrollLock",	state_scroll);
	
	/* Grrr, just found out that older glib doesn't have this useful function! */
	//g_key_file_save_to_file (config, filename, &error);
	contents = g_key_file_to_data (config, &length, NULL);
	g_file_set_contents (filename, contents, length, &error);
	
	if (error != NULL)
		g_printerr ("Error: Can't write config file!\n");

	g_key_file_free (config);
	g_free (filename);
}

/* ------------------------------------------------------------------------- */

void load_config (void) {

	gchar		*filename;
	GKeyFile	*config;
	GError		*error = NULL;

	filename = g_strconcat (g_get_user_config_dir(), "/", settings, NULL);

	config = g_key_file_new();
	g_key_file_load_from_file (config, filename, G_KEY_FILE_KEEP_COMMENTS, &error);

    if (error == NULL && g_key_file_has_group (config, "Settings")) {
	    state_caps		= g_key_file_get_boolean (config, "Settings", "CapsLock", NULL);
	    state_num		= g_key_file_get_boolean (config, "Settings", "NumLock", NULL);
	    state_scroll	= g_key_file_get_boolean (config, "Settings", "ScrollLock", NULL);
    } else {
        if (error) g_error_free (error);
		save_config();
	}

	g_key_file_free (config);
	g_free (filename);
}

/* ------------------------------------------------------------------------- */	

void toggle (GtkWidget *widget, gpointer data) {
	
	gint *state;
	GtkStatusIcon *icon;
	
	if (g_strcmp0( (gchar*)data, "caps") == 0) {
		icon = icon_caps;
		state = &state_caps;
	} else if (g_strcmp0( (gchar*)data, "num") == 0) {
		icon = icon_num;
		state = &state_num;
	} else {
		icon = icon_scroll;
		state = &state_scroll;
	}
	
	/* At least one icon has to be visible! */
	if ((state_caps + state_num + state_scroll <= 1) && *state) return;
	
	*state ^= 1;	/* toggle state */
	gtk_status_icon_set_visible (icon, *state);
	save_config();
}

/* ------------------------------------------------------------------------- */	

void show_about (GtkWidget *widget, gpointer data) {
	
	GtkWidget *dialog;
	gchar *message;
	
	message = g_strdup_printf ("<b>%s</b>\n\n%s", appname,
			  "Copyright © SFR 2017 (GNU GPL v3)\n\nSVG status icons created <b>by hand</b> (GNU GPL v2)");

	dialog = gtk_message_dialog_new_with_markup (NULL,
			 GTK_DIALOG_DESTROY_WITH_PARENT,
			 GTK_MESSAGE_OTHER,
			 GTK_BUTTONS_OK,
			 message);

	gtk_window_set_title (GTK_WINDOW (dialog), appname);
	gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_destroy (dialog);
	
	g_free (message);
}

/* ------------------------------------------------------------------------- */	

void popup_menu (GtkStatusIcon *status_icon, guint button, guint activate_time, gpointer user_data) {

	GtkWidget *menu;
	GtkWidget *separator;
	GtkWidget *menuitem_caps;
	GtkWidget *menuitem_num;
	GtkWidget *menuitem_scroll;
	GtkWidget *menuitem_about;
	GtkWidget *menuitem_quit;

	menu = gtk_menu_new();
	
	menuitem_caps = gtk_check_menu_item_new_with_label (_("CapsLock"));
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menuitem_caps), state_caps);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem_caps);
	g_signal_connect (G_OBJECT (menuitem_caps), "toggled", G_CALLBACK(toggle), (gpointer) "caps");

	menuitem_num = gtk_check_menu_item_new_with_label (_("NumLock"));
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menuitem_num), state_num);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem_num);
	g_signal_connect (G_OBJECT (menuitem_num), "toggled", G_CALLBACK(toggle), (gpointer) "num");
	
	menuitem_scroll = gtk_check_menu_item_new_with_label (_("ScrollLock"));
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menuitem_scroll), state_scroll);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem_scroll);
	g_signal_connect (G_OBJECT (menuitem_scroll), "toggled", G_CALLBACK(toggle), (gpointer) "scroll");
	
	separator = gtk_separator_menu_item_new();
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), separator);

	menuitem_about = gtk_image_menu_item_new_from_stock ("gtk-about", NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem_about);
	g_signal_connect (G_OBJECT (menuitem_about), "activate", G_CALLBACK(show_about), NULL);
		
	menuitem_quit = gtk_image_menu_item_new_from_stock ("gtk-quit", NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem_quit);
	g_signal_connect (G_OBJECT (menuitem_quit), "activate", G_CALLBACK(gtk_main_quit), NULL);
	
	gtk_widget_show_all (menu);
	gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL, button, activate_time);
}

/* ------------------------------------------------------------------------- */	

void update_icon (gint state, GtkStatusIcon *target, char *icon_on, char *icon_off, char *string) {
	
	gchar *buf;
	gchar *icon;

	switch (state) {
		case 0:
			icon = g_strdup_printf ("%s", icon_off);
			buf = g_strdup_printf ("%s %s", string, _("is OFF"));
			break;
		default:
			icon = g_strdup_printf ("%s", icon_on);
			buf = g_strdup_printf ("%s %s", string, _("is ON"));
			break;
	}

	if (g_file_test (icon, G_FILE_TEST_EXISTS)) {
		gtk_status_icon_set_from_file (target, icon);
	} else {
		gtk_status_icon_set_from_stock (target, "gtk-missing-image");
	}
	
	gtk_status_icon_set_tooltip_text (target, buf);

	g_free (buf);
	g_free (icon);
}

/* ------------------------------------------------------------------------- */	

gint on_timeout (gpointer data) {
	static gint state_caps_old		= -1;
	static gint state_num_old		= -1;
	static gint state_scroll_old	= -1;
	gint state_caps;
	gint state_num;
	gint state_scroll;
	XKeyboardState state;
	Display *display = XOpenDisplay (NULL); 
	XGetKeyboardControl (display, &state);
	XCloseDisplay (display);
	state_caps		= state.led_mask & 1;
	state_num		= state.led_mask & 2;
	state_scroll	= state.led_mask & 4;
	
	if (state_caps != state_caps_old) {
		update_icon (state_caps, icon_caps, caps_on, caps_off, _("CapsLock"));
		state_caps_old = state_caps;
	}
	if (state_num != state_num_old) {
		update_icon (state_num, icon_num, num_on, num_off, _("NumLock"));		
		state_num_old = state_num;
	}
	if (state_scroll != state_scroll_old) {
		update_icon (state_scroll, icon_scroll, scroll_on, scroll_off, _("ScrollLock"));				
		state_scroll_old = state_scroll;
	}
	return TRUE;
}

/* ------------------------------------------------------------------------- */
/* Workaround for LxQt-Panel in Fatdog:
   icons shouldn't be displayed at once, or else some/all of them will be broken.
   This delay makes it slightly less likely. */

gint on_init (gpointer data) {

	static gint cnt = 3;
	switch (cnt) {
		case 3:
			gtk_status_icon_set_visible (icon_caps, state_caps);
			break;
		case 2:
			gtk_status_icon_set_visible (icon_num, state_num);
			break;
		case 1:
			gtk_status_icon_set_visible (icon_scroll, state_scroll);
			break;
		default:
			return FALSE;	/* stops the timer */
			break;
	}
	--cnt;
	return TRUE;
}

/* ========================================================================= */

int main(int argc, char *argv[]) {  
	// mods for inline svg
	char* caps00 = svg;
	char* caps01 = rep_str(svg, "#bbbbbb", "#f0f0f0");
	caps01 = rep_str(caps01, ">a<", ">A<");
	char* num00 = rep_str(svg, ">a<", ">0<");
	char* num01 = rep_str(svg, ">a<", ">1<");
	num01 = rep_str(num01, "#bbbbbb", "#f0f0f0");
	char* scr00 = rep_str(svg, ">a<", ">↓<");
	scr00 = rep_str(scr00, "font-weight='normal'", "font-weight='bold'");
	char* scr01 = rep_str(scr00, "#bbbbbb", "#f0f0f0");
	fp = fopen(caps_off,"w+"); fputs(caps00,fp); fclose(fp);
	fp = fopen(caps_on,"w+"); fputs(caps01,fp); fclose(fp);
	fp = fopen(num_off,"w+"); fputs(num00,fp); fclose(fp);
	fp = fopen(num_on,"w+"); fputs(num01,fp); fclose(fp);
	fp = fopen(scroll_off,"w+"); fputs(scr00,fp); fclose(fp);
	fp = fopen(scroll_on,"w+"); fputs(scr01,fp); fclose(fp);
	
	setlocale (LC_ALL, "");
	bindtextdomain ("lockstate", "/usr/share/locale/");
	textdomain ("lockstate");
	bind_textdomain_codeset ("lockstate", "UTF-8");

	load_config();
	
	if (!state_caps && !state_num && !state_scroll) state_caps = TRUE;
	
	gtk_init (&argc, &argv);
	icon_caps = gtk_status_icon_new ();
	gtk_status_icon_set_visible (icon_caps, FALSE);
	g_signal_connect (G_OBJECT (icon_caps), "popup-menu", G_CALLBACK (popup_menu), NULL);
	icon_num = gtk_status_icon_new ();
	gtk_status_icon_set_visible (icon_num, FALSE);
	g_signal_connect (G_OBJECT (icon_num), "popup-menu", G_CALLBACK (popup_menu), NULL);
	icon_scroll = gtk_status_icon_new ();
	gtk_status_icon_set_visible (icon_scroll, FALSE);
	g_signal_connect (G_OBJECT (icon_scroll), "popup-menu", G_CALLBACK (popup_menu), NULL);
	gtk_timeout_add (100, on_init, NULL);		/* hack */
	gtk_timeout_add (100, on_timeout, NULL);
	gtk_main();
	return 0;
}

/* ========================================================================= */	
